#!/usr/bin/lua

-- wbgen2, (c) 2010 Tomasz Wlostowski/CERN BE-Co-HT
-- LICENSED UNDER GPL v2

-- File: cgen_c_headers.lua
--
-- The C header code generator.
--

function cgen_c_field_define(field, reg)
	local prefix=string.upper(periph.c_prefix).."_"..string.upper(reg.c_prefix).."_"..string.upper(field.c_prefix);
	
	emit("");
	emit("/* definitions for field: "..field.name.." in reg: "..reg.name.." */");

	if(field.type == BIT or field.type == MONOSTABLE)  then
		emit("#define "..prefix.." (1<<"..field.offset..")");
	else 
		print(field.offset, field.size);
		emit("#define "..prefix.."_MASK "..string.format("0x%08x", (math.pow(2, field.size)-1) * math.pow(2, field.offset)));
		emit("#define "..prefix.."_SHIFT "..string.format("%d", field.offset));
		emit("#define "..prefix.."_W(value) "..string.format("(((value) & 0x%08x) << %d)", math.pow(2, field.size)-1, field.offset));
		if(field.type == SIGNED) then
			emit("#define "..prefix.."_R(reg) "..string.format("( ((reg)&0x%08x ? 0x%08x : 0) | (((reg) >> %d) & 0x%08x))", 
																						math.pow(2, field.offset + field.size - 1),  -- sign mask
																						(math.pow(2, 32-field.size)-1) * math.pow(2, field.size),			 -- sign extension mask	
																						field.offset, 
																						math.pow(2, field.size)-1));
		else
			emit("#define "..prefix.."_R(reg) "..string.format("(((reg) >> %d) & 0x%08x)", field.offset, math.pow(2, field.size)-1));
		end
	end	
end

function cgen_c_ramdefs(ram)
	local prefix = string.upper(periph.c_prefix).."_"..string.upper(ram.c_prefix);

	emit("/* definitions for RAM: "..ram.name.." *);	

	emit(string.format("#define "..prefix.."_BYTES 0x%08x %-50s", ram.size * ram.width / 8, "/* size in bytes */"));
	emit(string.format("#define "..prefix.."_WORDS 0x%08x %-50s", ram.size, "/* size in "..ram.width.."-bit words, 32-bit aligned */"));
end

function cgen_c_field_masks()
	foreach_reg(function(reg)
								if(reg.__type == TYPE_REG and reg.num_fields ~= nil and reg.num_fields > 0) then
									emit("/* definitions for register: "..reg.name.." */);
									foreach_subfield(reg, function(field, reg) cgen_c_field_define(field, reg) end);
								elseif (reg.__type == TYPE_RAM) then
									cgen_c_ramdefs(reg);
								end
							end);
end

function cgen_c_headers_reg(reg)
	
	foreach_subfield(reg, function(field, reg)
													emit()
												end);
	
end

function cgen_c_headers_ram()

end


function cgen_c_fileheader()
	  emit ("/*");
    emit ("  Register definitions for slave core: "..periph.name);
    emit ("");
    emit ("  * File           : "..options.output_c_header_file);
    emit ("  * Author         : auto-generated by wbgen2 from "..input_wb_file);
    emit ("  * Created        : "..os.date());
    emit ("  * Standard       : ANSI C");
    emit ("");
    emit ("    THIS FILE WAS GENERATED BY wbgen2 FROM SOURCE FILE "..input_wb_file);
    emit ("    DO NOT HAND-EDIT UNLESS IT'S ABSOLUTELY NECESSARY!");
    emit ("");
    emit ("*/");
end

function cgen_c_struct()
	local cur_offset = 0;
	local pad_id = 0;
	
	function pad_struct(base)
		if(cur_offset < base) then
			emit("/* padding to: "..base.." words */");
			emit("uint32_t __padding_"..pad_id.."["..(base - cur_offset).."];");
			pad_id=pad_id+1;
			cur_offset = base;
		end
	end
	
	emit("PACKED struct "..string.upper(periph.c_prefix).."_WB {");
	indent_right();
	
	
	-- emit struct entires for REGs
	foreach_reg(function(reg)
								if(reg.__type == TYPE_REG) then
									pad_struct(reg.base);
									emit("/* "..reg.name.." */");
									emit("volatile uint32_t "..string.upper(reg.prefix)..";");
									cur_offset = cur_offset + 1;
								end
							end);

	-- .. and for RAMs
	foreach_reg(function(ram)
								if(ram.__type == TYPE_RAM) then
									local base = math.pow(2, ram.select_bits) * 
															 math.pow (2, address_bus_width - address_bus_select_bits);									
												
									pad_struct(base);									
									emiti();
									emitx("/* RAM: "..ram.name..", "..ram.size.." "..ram.width.."-bit words, "..DATA_BUS_WIDTH.."-bit aligned, "..csel(ram.byte_select, "byte", "word").."-addressable");
									
									if(ram.wrap_bits > 0) then
										emitx(", mirroring: "..math.pow(2, ram.wrap_bits).." times */\n");
									else
										emitx(" */\n");
									end
									
									if(ram.byte_select) then
										emit("volatile uint8_t "..string.upper(ram.prefix).." ["..(ram.size * (DATA_BUS_WIDTH/8) * math.pow(2, ram.wrap_bits)) .."];");
									else
										emit("volatile uint32_t "..string.upper(ram.prefix).." ["..(ram.size * math.pow(2, ram.wrap_bits)) .."];");									
									end
								end
							end);
	
					
	
	
	
	
	indent_left();
	emit("};");
end

function cgen_generate_c_header_code()

	cgen_new_snippet();
	
	cgen_c_fileheader();

	emit("");
	emit("#ifndef __WBGEN2_REGDEFS_"..string.upper(string.gsub(input_wb_file,"%.","_")))
	emit("#define __WBGEN2_REGDEFS_"..string.upper(string.gsub(input_wb_file,"%.","_")))
	emit("");
	emit("#include <inttypes.h>");
	emit("");	
	emit("#if defined( __GNUC__)");
	emit("#define PACKED __attribute__ ((packed))");
	emit("#else");
	emit("#error \"Unsupported compiler?\"");
	emit("#endif");
	emit("");
	emit("#ifndef __WBGEN2_MACROS_DEFINED__");
	emit("#define __WBGEN2_MACROS_DEFINED__");

	emit("#define WBGEN2_GENMASK(offset, size) (((1<<(size))-1) << (offset))");
	emit("#define WBGEN2_GENWRITE(reg, offset, size) (((reg) >> (offset)) & ((1<<(size))-1))");
	
	emit("#endif");

	cgen_c_field_masks();
	emit("");
	cgen_c_struct();
	emit("");	
	emit("#endif");
	cgen_write_current_snippet();
end
